#include <iostream.h>
#include <conio.h>
#include <string.h>
#include <stdlib.h>

template <class T>                       //a general template for class Vector for all fields
class Vector
{
	private: T* ptr; int size;            //T* contains members of Vector
   public:
   	Vector(int s)                      //within template class we can use simply Vector or Vector<T>
      {
      	size=s;
         if(size!=0)	ptr=new T[size];    //creating dynamic memory for given size in the constructor
         else	ptr=NULL;
      }
      Vector<T>(const Vector<T>& copy)   //copy constructor
      {
       	size=copy.size;
         if(size!=0)
         {
         	ptr=new T[size];             //creating memory dynamically for copy.size
            for(int i=0;i<size;i++)
            	ptr[i]=copy.ptr[i];
      	}
         else	ptr=NULL;
      }
      ~Vector<T>(){delete []ptr;}        //destructor deletes memory alocated dynamically

      const	Vector& operator=(const Vector<T>& right)
      {
      	if(this!=&right)                //overloaded operator '=' for user defined data type Vector
         {
         	delete[]ptr;
            size=right.size;
            if(size!=0)
            {
            	ptr=new T[size];
               for(int i=0;i<size;i++)
               	ptr[i]=right.ptr[i];
            }
            else	ptr=NULL;
         }
         return *this;
      }

      T& operator[](int index)   //overloaded operator '[]' for Vector
      {
      	if(index<0||index>=size)
         {	cout<<"Error: out of range\n"; exit(1);}
         return ptr[index];
      }

      void print()
      {
      	cout<<"template class vector is being traversed\n";
       	for(int i=0;i<size;i++)
         	cout<<ptr[i]<<'\t';
         cout<<endl;
      }
};

class Vector<char*> //template specialization for type 'char*' as general template can not handle 'char*' properly
{
	private:	char** ptr;	int size;  //ptr is array of pointers to characters so char**
   public:
   	Vector(int s)
      {
      	size=s;
         if(size!=0)
         {
         	ptr=new char*[size];       //creating dynamic memory for pointers
            for(int i=0;i<size;i++)
            	ptr[i]=0;               //setting all pointers to 0
         }
         else	ptr=NULL;
      }
      Vector(const Vector<char*>& copy)
      {                                  //copy constructor
      	size=copy.size;
         if(size==0)
         {ptr=NULL; return;}

         ptr=new char*[size];          //creating dynamic memory for pointers
         for(int i=0;i<size;i++)
         	if(copy.ptr[i]!=NULL)
            {
            	ptr[i]=new char[strlen(copy.ptr[i])+1]; //creating dynamic memories for all strings using loop
               strcpy(ptr[i],copy.ptr[i]);
            }
            else	ptr[i]=NULL;
      }
      ~Vector()
      {
      	for(int i=0;i<size;i++)     //deleting all memories of strings using loop
         	delete[]ptr[i];
         delete []ptr;
      }

      const char*& operator[](int index)    //overloaded operator '[]'
      {
       	if(index<0||index>=size)
         {	cout<<"Error: out of range\n"; exit(1);}
         return	ptr[index];
      }

      const Vector&	operator=(const Vector&right)    //overloaded '=' operator
      {
      	if(this==&right)
         	return *this;
         for(int i=0;i<size;i++)  //deleting previous strings to overcome with dynamic memory problems and dangling pointer issues
           	delete[]ptr[i];
         delete []ptr;            //deleting previous pointer
         size=right.size;
         if(size==0)
         {	ptr=NULL; return *this;}

         ptr=new char*[size];          //creating dynamic memory for pointer
         for(int i=0;i<size;i++)
         	if(right.ptr!=NULL)
            {
            	ptr[i]=new char[strlen(right.ptr[i])+1];
               strcpy(ptr[i],right.ptr[i]);
            }                                       //creating dynamic memory for all strings using loop
            else	ptr[i]=NULL;
      }

      void insert(char* str, int i)
      {                                 //insert fn inserts a string at ith index
      	delete [] ptr[i];
         if(str!=0)
         {
         	ptr[i]=new char[strlen(str)+1];
            strcpy(ptr[i],str);
         }
         else	ptr[i]=NULL;
      }

      void print()
      {
      	cout<<"specialized template class vector is being traversed\n";
         for(int i=0;i<size;i++)
         	cout<<ptr[i]<<'\t';
         cout<<endl;
      }
};

main()
{
	Vector<int>	  vi1(3),vi2(2);           //obj is being created using general template
   Vector<float> vf1(3),vf2(1);   			//general template
   Vector<char*> vc1(4),vc2(4);           //obj is being created using specialized template

   vi1[0]=3;	vi1[1]=4;	vi1[2]=5;
   vi2[0]=4;	vi2[1]=6;
   vi2=vi1;

   vf1[0]=9.5;	vf1[1]=3.5;	vf1[2]=4.5;
   vf2[0]=34.4;
   vf2=vf1;

   vc1[0]="ali";
   vc1.insert("waqas",3);	vc1.insert("DX",1);	vc1.insert("ahsan",2);
   vc2=vc1;

   Vector<int>   vi3=vi1;
   Vector<float> vf3=vf2;                //copy constructor is being called
   Vector<char*> vc3=vc1;

   vi1.print();	vi2.print();   vi3.print();	cout<<endl;
   vf1.print();	vf2.print();	vf3.print();	cout<<endl;
   vc1.print();	vc2.print();   vc3.print();	cout<<endl;

getch();
}


